/* glass : the gamecube linux awesome stream server
 * glass+ftc : extended version of glass, by FreeTheCube <free.the.cube@free.fr>.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <netinet/in.h>

#include <sys/socket.h>
#include <sys/types.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>
#include <signal.h>
#include <dirent.h>
#include <string.h>

#define PORT 5173
#define BUFFERSIZE 1028

//! The mount point.
static char mount_point[] = "/tmp/XXXXXX";

//! File descriptor of the game image.
static int fd_game = -1;

//! The new FST.
static uint8_t *fst_buffer = NULL;

/**
 * Build a new FST.
 * \param lock the path of the file which contain files to be locked, else NULL.
 * \return the new FST on succes, else NULL.
 */
uint8_t *create_new_fst(char *lock)
{
	int retcode = 0;
	uint32_t i = 0;
	uint32_t j = 0;
	uint32_t offset = 0;
	uint32_t fst_size = 0;
	uint32_t fst_offset = 0;
	FILE *fd = NULL;	
	uint8_t buffer[8];
	uint8_t *fst_buffer = NULL;
	char filename[256+1];

	fd = fopen(lock, "r");
	if(NULL == fd) {
		perror(lock);
		return NULL;
	}

	retcode = lseek(fd_game, 0x424, SEEK_SET);
	if(-1 == retcode) {
		perror("game image");
		return NULL;
	}
		
	retcode = read(fd_game, buffer, 8);
	if(-1 == retcode) {
		perror("game image");
		return NULL;
	}
	
	fst_offset = ((uint32_t *)buffer)[0];
	fst_offset = ntohl(fst_offset);
	fst_size = ((uint32_t *)buffer)[1];
	fst_size = ntohl(fst_size);

	// Allocate memory for the modified FST
	fst_buffer = (uint8_t *)calloc(fst_size, 1);
	if(NULL == fst_buffer) {
		perror(NULL);
		return NULL;
	}

	retcode = lseek(fd_game, fst_offset, SEEK_SET);
	if(-1 == retcode) {
		perror("game image");
		free(fst_buffer);
		return NULL;
	}

	retcode = read(fd_game, fst_buffer, fst_size);
	if(-1 == retcode) {
		perror("game image");
		free(fst_buffer);
		return NULL;
	}
	
	while(1) {
		retcode = fscanf(fd, "%d %*s\n", &offset);
		if(EOF == retcode)
			break;
		if(1 != retcode || offset >= fst_size) {
			fprintf(stderr, "%s: Bad format or out of range value\n", lock);
			free(fst_buffer);
			return NULL;
		}

		// Reverse the name of locked files, so they will not be findable.
		for(j = offset; fst_buffer[j] != 0; j++)
			filename[j - offset] = fst_buffer[j];

		j = j - offset;
		filename[j] = '\0';

		for(i = offset, j--; fst_buffer[i] != 0 || 0 == j; i++, j--)
			fst_buffer[i] = filename[j];
	}

	return fst_buffer;
}

/**
 * The stream server himself.
 * \param lock the path of the file which contain files to be locked, else NULL.
 */
void stream_server(char *lock)
{
	int i = 0;
	int j = 0;
	int k = 0;	
	int retcode = 0;

	uint32_t filesize = 0;
	uint32_t dataread = 0;
	uint32_t fileoffset = 0;
	uint32_t fst_offset = 0;

	int mysock = 0;
	int addrsize = 0;
	int numpackets = 0;
	
	char buffer[BUFFERSIZE];
	char active_bar[] = {'-', '\\', '|', '/'};

	struct sockaddr_in myaddr, gcaddr;

	bzero(&myaddr, sizeof(myaddr));
	myaddr.sin_family = AF_INET;
	myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	myaddr.sin_port = htons(PORT);

	mysock = socket(AF_INET, SOCK_DGRAM, 0);
	if(-1 == mysock) {
		perror(NULL);
		return;
	}

	retcode = bind(mysock, (struct sockaddr *)&myaddr, sizeof(myaddr));
	if(-1 == retcode) {
		perror(NULL);
		return;
	}
		
	if(NULL != lock) {
		fst_buffer = create_new_fst(lock);
		if(NULL != fst_buffer)
			printf("Lock files system enabled\n");
		else printf("Lock files system disabled\n");
	}

	retcode = lseek(fd_game, 0, SEEK_SET);
	if(-1 == retcode) {
		perror("game image");
		return;
	}

	retcode = read(fd_game, buffer, 256);
	if(-1 == retcode) {
		perror("game image");
		return;
	}

	buffer[255] = 0;
	printf("%c%c%c (", buffer[0], buffer[1], buffer[2]);
	switch(buffer[3]){
	case 'E':
		printf("USA/NTSC");
		break;
	case 'P':
		printf("Europe/PAL");
		break;		
	case 'J':
		printf("Japan/NTSC");
		break;
	default:
		printf("Other");
	}
	printf(") : %s\n", buffer+32);
	
	retcode = lseek(fd_game, 0x424, SEEK_SET);
	if(-1 == retcode) {
		perror("game image");
		return;
	}
		
	retcode = read(fd_game, buffer, 8);
	if(-1 == retcode) {
		perror("game image");
		return;
	}
	
	fst_offset = ((uint32_t *)buffer)[0];
	fst_offset = ntohl(fst_offset);

	k = 0;
	while(1) {
		bzero(&gcaddr, sizeof(gcaddr));
		addrsize = sizeof(gcaddr);
		recvfrom(mysock, buffer, BUFFERSIZE, 0, (struct sockaddr *)&gcaddr, &addrsize);
		
		// Display an active bar, to see if the game is crashed or not
		printf("\r%c", active_bar[k]);
		fflush(stdout);
		k = (k+1)%4;

		// The famous "Save Hack"
		if(((int *)buffer)[0] != 0 &&
		   1632116837 == ((int *)buffer)[1] &&
		   1159752547 == ((int *)buffer)[2]) {
			printf("\rSave hack enabled\n");
			fflush(stdout);
			continue;
		}

		fileoffset = ((uint32_t *)buffer)[1];
		fileoffset = ntohl(fileoffset);
		filesize = ((uint32_t *)buffer)[2];
		filesize = ntohl(filesize);
		numpackets = (filesize/(BUFFERSIZE-4))+1;
    
		if(fileoffset == fst_offset && NULL != fst_buffer) {
			// Send the modified FST
			for(i = 0; i < numpackets; i++) {
				if(i+1 == numpackets)
					dataread = filesize%(BUFFERSIZE-4);
				else
					dataread = BUFFERSIZE-4;

				for(j = 0; j < dataread; j++)
					buffer[j+4] = fst_buffer[i*(BUFFERSIZE-4)+j];

				buffer[0] = 1;
				buffer[2] = i/256;
				buffer[3] = i%256;

				sendto(mysock, buffer, dataread+4, 0, (struct sockaddr *)&gcaddr, addrsize);
			}
			//end of stream
			buffer[0] = 2;
			sendto(mysock, buffer, 4, 0, (struct sockaddr *)&gcaddr, addrsize);
			
			continue;
		}

		lseek(fd_game, fileoffset, SEEK_SET);

		for(i = 0; i < numpackets; i++) {
			if(i+1 == numpackets) //last packet
				dataread = read(fd_game, buffer+4, filesize%(BUFFERSIZE-4));
			else
				dataread = read(fd_game, buffer+4, BUFFERSIZE-4);
			buffer[0] = 1;
			buffer[2] = i/256;
			buffer[3] = i%256;

			sendto(mysock, buffer, dataread+4, 0, (struct sockaddr *)&gcaddr, addrsize);
		}
		//end of stream
		buffer[0] = 2;
		sendto(mysock, buffer, 4, 0, (struct sockaddr *)&gcaddr, addrsize);
	}

	close(mysock);
}

/**
 * Open the game file.
 * Automaticly detect the format (iso/ziso).
 * \param filename the name of the game file.
 * \return the file descriptor of the game, else -1.
 */
int open_game(char *filename)
{
	int fd = 0;
	int retcode = 0;
	uint8_t buffer[BUFFERSIZE];
	uint8_t command[1024+1];
	DIR *dir = NULL;
	struct dirent *dirent = NULL;
	char *file_game = NULL;

	fd = open(filename, O_RDONLY);
	if(-1 == fd){
		perror(filename);
		return -1;
	}
	
	retcode = lseek(fd, 32808, SEEK_SET);
	if(-1 == retcode) {
		perror(filename);
		return -1;
	}

	retcode = read(fd, buffer, 5);
	if(-1 == retcode) {
		perror(filename);
		return -1;
	}

	//TODO to be improved
	if(!strncmp("CDROM", buffer, 5)) {
		printf("Compressed ISO9660 File System detected\n");
		if(NULL == mkdtemp(mount_point)) {
			perror(mount_point);
			return -1;
		}

		//TODO to be improved
		memset(command, 0, 1024+1);
		snprintf(command, 1024, "mount -o loop %s %s", filename, mount_point);
		retcode = system(command);
		if(retcode != 0) {
			fprintf(stderr, "unable to mount the file system\n");
			return -1;
		}
		
		dir = opendir(mount_point);
		if(NULL == dir) {
			perror(mount_point);
			return -1;
		}

		//TODO to be improved
		dirent = readdir(dir); //.
		dirent = readdir(dir); //..
		dirent = readdir(dir); //the game
		if(NULL == dirent) {
			perror(mount_point);
			return -1;
		}
		
		memset(command, 0, 1024+1);
		snprintf(command, 1024, "%s/%s", mount_point, dirent->d_name);

		retcode = closedir(dir);
		if(-1 == retcode) {
			perror(mount_point);
			return -1;
		}

		file_game = command;
	}
	else file_game = filename;

	fd = open(file_game, O_RDONLY);
	if(-1 == fd) {
		perror(filename);
		return -1;
	}

	return fd;
}

/**
 * Terminat properly the program.
 * Close all open resources.
 * \param the sigan lnumber.
 */
void handler(int i)
{
	int retcode = 0;
	char command[1024+1];

	fprintf(stderr, "\rExiting\n");

	if(NULL != fst_buffer)
		free(fst_buffer);

	if(-1 != fd_game) {
		retcode = close(fd_game);
		if(-1 == retcode) {
			perror("game image");
			return;
		}
	}

	if(strncmp(mount_point, "/tmp/XXXXXX", 11)) {
		//TODO to be improved
		snprintf(command, 1024, "umount %s", mount_point);
		retcode = system(command);
		if(retcode != 0) {
			perror(mount_point);
			return;
		}
		
		retcode = rmdir(mount_point);
		if(-1 == retcode) {
			perror(mount_point);
			return;
		}
	}

	exit(1);
}

/**
 * Start entry of the program.
 * Install the signal handler.
 * \param argc the number of arguments.
 * \param argv the array of arguments.
 * \return never
 */
int main(int argc,char **argv)
{
	int retcode = 0;
	struct sigaction sig_action;
	char *lock = NULL;


	switch(argc) {
	case 3:
		lock = argv[2];
		break;
	case 2:
		lock = NULL;
		break;
	default:
		fprintf(stderr, "usage : %s <filename.[z]iso> [filename.lock]\n", argv[0]);
		exit(1);
	}

	sig_action.sa_handler = handler;
	retcode = sigemptyset(&(sig_action.sa_mask));
	if(-1 == retcode)
		fprintf(stderr, "sigemptyset failed\n");
	else {
		sig_action.sa_flags = 0;
		retcode = sigaction(SIGHUP, &sig_action, NULL);
		retcode |= sigaction(SIGINT, &sig_action, NULL);
		retcode |= sigaction(SIGPIPE, &sig_action, NULL);
		retcode |= sigaction(SIGALRM, &sig_action, NULL);
		retcode |= sigaction(SIGTERM, &sig_action, NULL);
		retcode |= sigaction(SIGUSR1, &sig_action, NULL);
		retcode |= sigaction(SIGUSR2, &sig_action, NULL);
		retcode |= sigaction(SIGQUIT, &sig_action, NULL);
		retcode |= sigaction(SIGILL, &sig_action, NULL);
		retcode |= sigaction(SIGABRT, &sig_action, NULL);
		retcode |= sigaction(SIGFPE, &sig_action, NULL);
		retcode |= sigaction(SIGSEGV, &sig_action, NULL);

		if(retcode != 0)
			fprintf(stderr, "sigaction failed\n");
	}

	fd_game = open_game(argv[1]);
	if(-1 != fd_game)
		stream_server(lock);
	
	handler(0);

	return -1;
}
